@@ -34,6 +34,7 @@ typedef struct {
     PyObject_HEAD
     /* Type-specific fields go here. */
     TMR_Reader reader;
+    TMR_TagFilter tag_filter;
     TMR_TagOp tagop;
     uint8_t antennas[MAX_ANTENNA_COUNT];
     TMR_ReadListenerBlock readListener;
@@ -198,12 +199,15 @@ Reader_set_read_plan(Reader *self, PyObject *args, PyObject *kwds)
     TMR_Status ret;
     int i;
     uint8_t ant_count;
+    char* epc_target = NULL;
+    int target_len;
+    TMR_TagData target;
     PyObject *bank = NULL;
     int readPower = 0;
 
-    static char *kwlist[] = {"antennas", "protocol", "bank", "read_power", NULL};
+    static char *kwlist[] = {"antennas", "protocol", "epc_target", "bank", "read_power", NULL};
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!s|Oi", kwlist, &PyList_Type, &list, &s, &bank, &readPower))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!s|z#Oi", kwlist, &PyList_Type, &list, &s, &epc_target, &target_len, &bank, &readPower))
         return NULL;
 
     if ((protocol = str2protocol(s)) == TMR_TAG_PROTOCOL_NONE)
@@ -233,6 +237,19 @@ Reader_set_read_plan(Reader *self, PyObject *args, PyObject *kwds)
     if ((ret = TMR_RP_init_simple(&plan, ant_count, self->antennas, protocol, 1000)) != TMR_SUCCESS)
         goto fail;
 
+    if(epc_target != NULL)
+    {
+        target.epcByteCount = target_len/2;
+        TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
+
+        if ((ret = TMR_TF_init_tag(&self->tag_filter, &target)) != TMR_SUCCESS)
+            goto fail;
+
+        TMR_RP_set_filter(&plan, &self->tag_filter);
+    }
+    else
+        TMR_RP_set_filter(&plan, NULL);
+
     if (bank != NULL)
     {
         int op = 0;
@@ -258,9 +275,10 @@ Reader_set_read_plan(Reader *self, PyObject *args, PyObject *kwds)
         if ((ret = TMR_TagOp_init_GEN2_ReadData(&self->tagop, op, 0, 0)) != TMR_SUCCESS)
             goto fail;
 
-        if ((ret = TMR_RP_set_tagop(&plan, &self->tagop)) != TMR_SUCCESS)
-            goto fail;
+        TMR_RP_set_tagop(&plan, &self->tagop);
     }
+    else
+        TMR_RP_set_tagop(&plan, NULL);
 
     if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_READ_PLAN, &plan)) != TMR_SUCCESS)
         goto fail;
@@ -282,6 +300,7 @@ Reader_write(Reader *self, PyObject *args, PyObject *kwds)
 {
     char* epc_data;
     char* epc_target = NULL;
+    int data_len, target_len;
     TMR_Status ret;
     TMR_TagData data;
     TMR_TagData target;
@@ -289,17 +308,17 @@ Reader_write(Reader *self, PyObject *args, PyObject *kwds)
 
     // Read call arguments.
     static char *kwlist[] = {"epc_code", "epc_target", NULL};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|s", kwlist, &epc_data, &epc_target))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s#|z#", kwlist, &epc_data, &data_len, &epc_target, &target_len))
         return NULL;
 
     /* build data tag to be writen */
-    data.epcByteCount = strlen(epc_data) * sizeof(char) / 2;
+    data.epcByteCount = data_len/2;
     TMR_hexToBytes(epc_data, data.epc, data.epcByteCount, NULL);
 
     /* build target tag to search */
     if(epc_target != NULL)
     {
-        target.epcByteCount = strlen(epc_target) * sizeof(char) / 2;
+        target.epcByteCount = target_len/2;
         TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
 
         filter = &tag_filter;
@@ -508,6 +527,7 @@ Reader_read_tag_mem(Reader *self, PyObject *args, PyObject *kwds)
 {
     TMR_Status ret;
     char* epc_target = NULL;
+    int target_len;
     uint32_t bank, address, count;
 
     TMR_TagData target;
@@ -516,13 +536,13 @@ Reader_read_tag_mem(Reader *self, PyObject *args, PyObject *kwds)
     PyObject *result;
 
     static char *kwlist[] = {"bank", "address", "count", "epc_target", NULL};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "III|s", kwlist, &bank, &address, &count, &epc_target))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "III|z#", kwlist, &bank, &address, &count, &epc_target, &target_len))
         return NULL;
 
     if(epc_target != NULL)
     {
         /* build target tag to search */
-        target.epcByteCount = strlen(epc_target) / 2;
+        target.epcByteCount = target_len/2;
         TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
 
         filter = &tag_filter;
@@ -553,6 +573,7 @@ Reader_write_tag_mem(Reader *self, PyObject *args, PyObject *kwds)
 {
     TMR_Status ret;
     char* epc_target = NULL;
+    int target_len;
     uint32_t bank, address;
     PyObject *data;
 
@@ -560,13 +581,13 @@ Reader_write_tag_mem(Reader *self, PyObject *args, PyObject *kwds)
     TMR_TagFilter tag_filter, *filter;
 
     static char *kwlist[] = {"bank", "address", "data", "epc_target", NULL};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "IIO!|s", kwlist, &bank, &address, &PyByteArray_Type, &data, &epc_target))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "IIO!|z#", kwlist, &bank, &address, &PyByteArray_Type, &data, &epc_target, &target_len))
         return NULL;
 
     if(epc_target != NULL)
     {
         /* build target tag to search */
-        target.epcByteCount = strlen(epc_target) / 2;
+        target.epcByteCount = target_len/2;
         TMR_hexToBytes(epc_target, target.epc, target.epcByteCount, NULL);
 
         filter = &tag_filter;
@@ -658,22 +679,35 @@ Reader_gpo_set(Reader *self, PyObject *args)
 /* Functions to get/set the reader parameters */
 
 static PyObject *
-Reader_get_model(Reader* self)
+get_string(TMR_Reader *reader, int param)
 {
-    TMR_String model;
-    char str[64];
+    TMR_String str;
+    char buffer[64];
     TMR_Status ret;
 
-    model.value = str;
-    model.max = sizeof(str);
+    str.value = buffer;
+    str.max = sizeof(buffer);
 
-    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_VERSION_MODEL, &model)) != TMR_SUCCESS)
+    if ((ret = TMR_paramGet(reader, param, &str)) != TMR_SUCCESS)
     {
-        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(&self->reader, ret));
+        PyErr_SetString(PyExc_RuntimeError, TMR_strerr(reader, ret));
         return NULL;
     }
 
-    return PyUnicode_FromString(model.value);
+    return PyUnicode_FromString(str.value);
+}
+
+
+static PyObject *
+Reader_get_model(Reader* self)
+{
+    return get_string(&self->reader, TMR_PARAM_VERSION_MODEL);
+}
+
+static PyObject *
+Reader_get_serial(Reader* self)
+{
+    return get_string(&self->reader, TMR_PARAM_VERSION_SERIAL);
 }
 
 typedef struct {
@@ -1418,6 +1452,9 @@ static PyMethodDef Reader_methods[] = {
     {"get_model", (PyCFunction)Reader_get_model, METH_NOARGS,
      "Returns the model name"
     },
+    {"get_serial", (PyCFunction)Reader_get_serial, METH_NOARGS,
+     "Returns a serial number of the reader, the same number printed on the barcode label"
+    },
     {"set_region", (PyCFunction)Reader_set_region, METH_VARARGS,
      "Set the reader region"
     },
